"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = require("path");
const broccoli_debug_1 = __importDefault(require("broccoli-debug"));
const broccoli_funnel_1 = __importDefault(require("broccoli-funnel"));
const broccoli_merge_trees_1 = __importDefault(require("broccoli-merge-trees"));
const broccoli_source_1 = require("broccoli-source");
const lodash_frompairs_1 = __importDefault(require("lodash.frompairs"));
const commands_1 = __importDefault(require("./commands"));
const options_1 = require("./lib/options");
const ember_cli_entities_1 = require("./lib/utils/ember-cli-entities");
const config_creator_1 = require("./plugins/broccoli/config-creator");
const ember_css_modules_1 = __importDefault(require("./plugins/ember-css-modules"));
const preprocessor_registry_1 = require("./plugins/preprocessor-registry");
const addonPrototype = ember_cli_entities_1.addon({
    name: require(`${__dirname}/../package`).name,
    makeupOptions: undefined,
    themePackages: undefined,
    parentAddon: undefined,
    get usages() {
        return {};
    },
    get debugTree() {
        return broccoli_debug_1.default.buildDebugCallback(this.name);
    },
    findThemePackages() {
        const keyword = 'ember-makeup-theme';
        const { dependencies = {}, devDependencies = {} } = this.project.pkg;
        const packages = [dependencies, devDependencies]
            .flatMap(Object.keys)
            .map(name => this.project.require(`${name}/package.json`))
            .filter(json => json.keywords && json.keywords.includes(keyword))
            .map(json => ({
            name: (json.makeup && json.makeup.name) || json.name,
            package: json
        }));
        if (packages.length === 0) {
            throw new Error(`Could not find a package with the '${keyword}' keyword.`);
        }
        return packages;
    },
    included(includer) {
        this.computeOptions(includer);
        this.themePackages = this.findThemePackages();
        this._super.included.call(this, includer);
    },
    includedCommands() {
        return commands_1.default;
    },
    computeOptions(includer) {
        if (this.makeupOptions)
            return;
        if (!includer.options && !this.app)
            throw new Error('Could not find parent options.');
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        const parentOptions = includer.options || this.app.options;
        this.makeupOptions = options_1.computeOptions(parentOptions.makeup);
        if (this.parent.parent) {
            this.parentAddon = includer;
        }
    },
    shouldIncludeChildAddon(childAddon) {
        const disabledAddons = ['ember-css-modules', 'ember-cli-sass'];
        if (disabledAddons.includes(childAddon.name)) {
            return false;
        }
        return this._super.shouldIncludeChildAddon.call(this, childAddon);
    },
    /**
     * Integrate with other CSS processors, like ember-cli-sass.
     *
     * @see https://github.com/ember-cli/ember-cli-preprocess-registry#addon-usage
     */
    setupPreprocessorRegistry(type, registry) {
        preprocessor_registry_1.register(this, type, registry);
    },
    /**
     * Integrate with ember-css-modules.
     *
     * @see https://github.com/salsify/ember-css-modules/blob/master/docs/PLUGINS.md#plugins
     */
    createCssModulesPlugin(parent) {
        this.computeOptions(parent);
        return new ember_css_modules_1.default(parent, this);
    },
    reportUsages(callsite, usages) {
        this.usages[callsite] = usages;
    },
    treeForAddon(tree) {
        const originalTree = this.debugTree(this._super.treeForAddon.call(this, tree), 'treeForAddon:input');
        // Only run for the root app.
        if (this.parentAddon)
            return originalTree;
        const configFile = config_creator_1.configCreatorJS(`${this.name}/config`, {
            options: this.makeupOptions,
            themePaths: lodash_frompairs_1.default(this.themePackages.map(theme => [
                theme.name,
                path_1.join('/', this.makeupOptions.pathPrefix, `${theme.name}.css`)
            ]))
        });
        const mergedTree = this.debugTree(new broccoli_merge_trees_1.default([originalTree, configFile], {
            annotation: 'ember-makeup:merge-config'
        }), 'treeForAddon:output');
        return mergedTree;
    },
    treeForPublic() {
        // Only run for the root app.
        if (this.parentAddon)
            return undefined;
        const themePackageNames = this.findThemePackages();
        const themePackageSourceNodes = themePackageNames.map(themePackage => [
            themePackage.name,
            new broccoli_source_1.WatchedDir(path_1.dirname(this.project.resolveSync(`${themePackage.package.name}/package.json`)), { annotation: `ember-makeup:theme-source (${themePackage.name})` })
        ]);
        return this.debugTree(new broccoli_funnel_1.default(new broccoli_merge_trees_1.default(themePackageSourceNodes.map(([themeName, source]) => config_creator_1.configCreatorCSS(source, {
            themeName,
            customPropertyPrefix: this.makeupOptions.customPropertyPrefix,
            contextClassNamePrefix: this.makeupOptions.contextClassNamePrefix
        }))), { destDir: this.makeupOptions.pathPrefix }), 'treeForPublic:output');
    }
});
exports.default = addonPrototype;
exports.addonName = addonPrototype.name;
