"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const postcss_value_parser_1 = __importDefault(require("postcss-value-parser"));
const config_key_1 = require("../../lib/config-key");
const utils_1 = require("./utils");
function buildNamespaceFromParents(componentAssociations, node) {
    const namespaceFragments = [];
    for (const parent of utils_1.getParents(node)) {
        if (!componentAssociations.has(parent))
            continue;
        // @todo https://github.com/microsoft/TypeScript/issues/13086
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        const currentNamespaceFragment = componentAssociations.get(parent);
        if (config_key_1.isValidRootConfigKey(currentNamespaceFragment)) {
            namespaceFragments.push(config_key_1.resolveRootConfigKey(currentNamespaceFragment));
            break;
        }
        namespaceFragments.push(currentNamespaceFragment);
    }
    return namespaceFragments.reverse().join('.');
}
function namespaceConfigKey(namespace, key) {
    if (config_key_1.isValidRootConfigKey(key))
        return config_key_1.resolveRootConfigKey(key);
    return `${namespace}.${key}`;
}
function transformContextAtRule(atRule, namespace) {
    const configKey = atRule.params.length === 0 ? 'context' : atRule.params;
    utils_1.assertValidConfigKey(configKey, atRule);
    atRule.params = namespaceConfigKey(namespace, configKey);
}
function transformDecl(decl, namespace, { configKeyword, reportUsage, path }) {
    const originalValue = decl.value;
    decl.value = postcss_value_parser_1.default(originalValue)
        .walk(node => {
        if (node.type !== 'function' || node.value !== configKeyword)
            return;
        const keyNode = utils_1.getKeyNodeFromFunctionNode(decl, node, configKeyword);
        const key = namespaceConfigKey(namespace, keyNode.value);
        keyNode.value = key;
        if (reportUsage) {
            const rule = utils_1.findInIterable(utils_1.getParents(decl), parentNode => parentNode.type === 'rule');
            reportUsage({
                selectors: rule ? rule.selectors : [],
                property: decl.prop,
                originalValue,
                key,
                path
            });
        }
    })
        .toString();
}
function transformOne(container, component, { configKeyword, reportUsage, path }) {
    container.walkDecls(decl => {
        // Only parse and transform, if the function, e.g. `cfg(`, is present.
        if (!decl.value.includes(`${configKeyword}(`))
            return;
        transformDecl(decl, component, { configKeyword, reportUsage, path });
    });
    container.walkAtRules('context', atRule => transformContextAtRule(atRule, component));
}
function transformMany(container, componentAssociations, { configKeyword = 'cfg', reportUsage, path }) {
    container.walkDecls(decl => {
        // Only parse and transform, if the function, e.g. `cfg(`, is present.
        if (!decl.value.includes(`${configKeyword}(`))
            return;
        const parentNamespace = buildNamespaceFromParents(componentAssociations, decl);
        if (!parentNamespace)
            return;
        transformDecl(decl, parentNamespace, { configKeyword, reportUsage, path });
    });
    container.walkAtRules('context', atRule => {
        const parentNamespace = buildNamespaceFromParents(componentAssociations, atRule);
        transformContextAtRule(atRule, parentNamespace);
    });
}
/**
 * Expands `@component` at-rule shortcuts.
 *
 * ```css
 * @component button;
 *
 * .primary-button {
 *   @component primary;
 *   background: cfg('background');
 *   font-size: cfg('/typo.sizes.medium');
 * }
 *
 * .something-else {
 *   @component /typo.sizes;
 *  font-size: cfg('medium');
 * }
 * ```
 *
 * ```css
 * .primary-button {
 *   background: cfg('button.primary.background');
 *   font-size: cfg('typo.sizes.medium');
 * }
 *
 * .something-else {
 *   font-size: cfg('typo.sizes.medium');
 * }
 * ```
 */
exports.default = utils_1.pluginWithRequiredOptions('postcss-ember-makeup:expand-component-shortcut', function ({ configKeyword = 'cfg', componentKeyword = 'component', reportUsage, to }) {
    // eslint-disable-next-line unicorn/prevent-abbreviations
    return function (root, { opts }) {
        const path = to || (opts && opts.relativeFrom);
        if (!path)
            throw new TypeError('Neither `to` or `relativeFrom` were defined.');
        utils_1.assertNoDoubleColonAtRule(componentKeyword, root);
        const componentAssociations = new Map();
        root.walkAtRules(componentKeyword, atRule => {
            const { params: configKey, parent } = atRule;
            utils_1.assertValidConfigKey(configKey, atRule);
            if (componentAssociations.has(parent)) {
                throw atRule.error(`You can only specify a '@component' once per container.`);
            }
            componentAssociations.set(parent, configKey);
            atRule.remove();
        });
        switch (componentAssociations.size) {
            case 0:
                return;
            case 1: {
                const [container, component] = componentAssociations.entries().next().value;
                transformOne(container, component, {
                    configKeyword,
                    reportUsage,
                    path
                });
                break;
            }
            default:
                transformMany(root, componentAssociations, {
                    configKeyword,
                    reportUsage,
                    path
                });
        }
    };
});
