"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const postcss_1 = require("postcss");
const postcss_value_parser_1 = require("postcss-value-parser");
const config_key_1 = require("../../lib/config-key");
function assertNoDoubleColonAtRule(keyword, container) {
    container.walkAtRules(`${keyword}:`, atRule => {
        throw atRule.error(`Use '@${keyword}' without a ':'.`, { word: ':' });
    });
}
exports.assertNoDoubleColonAtRule = assertNoDoubleColonAtRule;
function assertValidConfigKey(configKey, container) {
    if (!config_key_1.isValidConfigKey(configKey)) {
        if (config_key_1.containsQuotes(configKey)) {
            throw container.error(`Do not put the config key in quotes.`, {
                word: configKey
            });
        }
        throw container.error(`Invalid config key: '${configKey}'`, {
            word: configKey
        });
    }
}
exports.assertValidConfigKey = assertValidConfigKey;
/**
 * Validate and return then key node from a parsed function node.
 * This ensures that the passed in function node only has a single parameter.
 */
function getKeyNodeFromFunctionNode(decl, node, keyword = 'cfg') {
    const [keyNode] = node.nodes;
    /**
     * When invoked as `cfg()`, without an explicit config key, return the
     * declaration property name as the key. For instance, the following would
     * return `border-color`:
     *
     * ```css
     * .foo {
     *   border-color: cfg();
     * }
     * ```
     */
    if (!keyNode) {
        const syntheticKeyNode = {
            type: 'string',
            value: decl.prop,
            quote: "'",
            unclosed: false
        };
        node.nodes = [syntheticKeyNode];
        return syntheticKeyNode;
    }
    /**
     * Disallow passing more than one parameter.
     */
    if (node.nodes.length !== 1) {
        throw decl.error(`You can only pass one parameter to '${keyword}', but you provided: ${postcss_value_parser_1.stringify(node)}`, { word: keyword });
    }
    /**
     * Only allow strings or words as keys.
     *
     * - string: `cfg('foo')`
     * - word:   `cfg(foo)`
     */
    if ((keyNode.type !== 'word' && keyNode.type !== 'string') ||
        keyNode.value.length === 0) {
        throw decl.error(`The first parameter of '${keyword}' has to be a key, but you provided: ${postcss_value_parser_1.stringify(node)}`, { word: keyword });
    }
    return keyNode;
}
exports.getKeyNodeFromFunctionNode = getKeyNodeFromFunctionNode;
function pluginWithRequiredOptions(name, initializer) {
    return postcss_1.plugin(name, (pluginOptions) => {
        if (!pluginOptions)
            throw new TypeError(`Plugin '${name}' requires you to pass an options object.`);
        return initializer(pluginOptions);
    });
}
exports.pluginWithRequiredOptions = pluginWithRequiredOptions;
function* getParents(node) {
    let { parent } = node;
    while (parent) {
        yield parent;
        parent = parent.parent;
    }
}
exports.getParents = getParents;
function findInIterable(iterable, predicate) {
    for (const element of iterable) {
        if (predicate(element))
            return element;
    }
    return null;
}
exports.findInIterable = findInIterable;
