"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const lodash_transform_1 = __importDefault(require("lodash.transform"));
const config_key_1 = require("../lib/config-key");
const isObject = (subject) => subject instanceof Object && !Array.isArray(subject);
const isThemeValue = (value) => typeof value === 'string' || typeof value === 'number';
const buildKey = (keyPrefix, key) => [...keyPrefix, key].join('.');
const createKeyError = (keyPrefix, key, message) => new TypeError(`'${buildKey(keyPrefix, key)}': ${message}`);
function isThemeSource(themeSource) {
    if (!isObject(themeSource))
        return false;
    if (!isObject(themeSource.components))
        return false;
    return true;
}
exports.isThemeSource = isThemeSource;
/* eslint-disable no-param-reassign */
const flattenTheme = (options, keyPrefix = [], acc = {
    contextless: {},
    contextual: {}
}) => (theme) => lodash_transform_1.default(theme, ({ contextless, contextual }, themeNode, key) => {
    if (typeof themeNode === 'object') {
        if (!themeNode || Array.isArray(themeNode)) {
            throw createKeyError(keyPrefix, key, '');
        }
        const nestedEntries = Object.entries(themeNode);
        if (nestedEntries.length === 0) {
            throw createKeyError(keyPrefix, key, 'Empty nested config objects are not allowed.');
        }
        const contexts = nestedEntries.filter(([nestedKey]) => config_key_1.isContextKey(nestedKey));
        if (contexts.length !== 0 && contexts.length !== nestedEntries.length) {
            throw createKeyError(keyPrefix, key, 'Mixing contextual keys and regular keys is not allowed.');
        }
        if (contexts.length > 0) {
            for (const [contextKey, value] of contexts) {
                if (!isThemeValue(value)) {
                    throw createKeyError(keyPrefix, key, 'Contextual configs must be primitive values.');
                }
                const contextName = config_key_1.contextFromContextKey(contextKey);
                if (!contextual[contextName]) {
                    contextual[contextName] = {};
                }
                contextual[contextName][buildKey(keyPrefix, key)] = options.processValue ? options.processValue(value) : value;
            }
            return;
        }
        flattenTheme(options, [...keyPrefix, key], {
            contextless,
            contextual
        })(themeNode);
        return;
    }
    if (isThemeValue(themeNode)) {
        contextless[buildKey(keyPrefix, key)] = options.processValue
            ? options.processValue(themeNode)
            : themeNode;
        return;
    }
    throw createKeyError(keyPrefix, key, `Unsupported value type '${themeNode}' (${typeof themeNode}).`);
}, acc);
/* eslint-enable no-param-reassign */
const makeInterpolator = (tokens) => (value) => typeof value === 'string'
    ? value.replace(/\$\{([^}]+)\}|\$([^\s]+)/g, (_, withCurly, withoutCurly) => {
        const id = withCurly || withoutCurly;
        if (!(id in tokens))
            throw new TypeError(`Token '${id}' is unknown.`);
        return String(tokens[id]);
    })
    : value;
function processTheme(themeSource) {
    const tokens = themeSource.tokens
        ? flattenTheme({ allowContexts: false })(themeSource.tokens).contextless
        : {};
    const interpolateWithTokens = makeInterpolator(tokens);
    const components = flattenTheme({
        allowContexts: true,
        processValue: interpolateWithTokens
    })(themeSource.components);
    return components;
}
exports.processTheme = processTheme;
